Require Export List.
Require Export Term_const.
Require Export pmov.
Require Export fpmov.
Require Export fpmovaux.
Require Omega.
Require Export Wf_nat.


 Lemma f2ind:
 forall (S1 S2 : State),
 (forall (l : Moves),  (S1 <> (nil, nil, l))) ->
 stepf S1 = S2 ->  dstep tmp S1 S2.
intros S1 S2 Hneq; destruct S1 as [[t b] l].
destruct b.
destruct t.
(* cas nil nil *)
elim (Hneq l); auto.
(* cas t nil *)
destruct m.
simpl; case (eq_nat_dec r r0); simpl.
(* nop *)
intros e H3; rewrite <- H3.
rewrite e; auto.
generalize dstep_nop.
unfold S; auto.
(* start *)
generalize (dstep_start tmp t l r r0); unfold S; auto.
intros H3 n H4; rewrite <- H4; auto.
(* t , cons *)
intros H; rewrite <- H; destruct m as [s d].
CaseEq (split_move t d).
(* cas Push *)
intros p H0; destruct p as [[t1 s0] t2].
simpl; rewrite H0.
destruct t; simpl.
inversion H0.
rewrite (unsplit_move (m :: t) t1 t2 d s0); auto.
destruct m.
generalize (dstep_push tmp); unfold S; intros H1; apply H1.
apply (splitSome ((r, r0) :: t) t1 t2 d s0); auto.
intros H0.
destruct b.
(* cas Last *)
simpl; rewrite H0.
destruct t; (try destruct m); generalize dstep_last; unfold S; intros H1;
 apply H1; apply splitNone; assumption.
(*cas Pops*)
rewrite (rebuild_l b m).
destruct (last (m :: b)).
case (eq_nat_dec d r).
(*cas PopLoop*)
intros e; rewrite <- e.
CaseEq (head_but_last (m :: b)); intros; [simpl | idtac];
 (try
   (destruct t; (try destruct m0); rewrite H0;
     (case (eq_nat_dec d d); intros h; (try (elim h; auto))))).
generalize (dstep_pop_loop tmp nil nil); unfold S; simpl; intros H2; apply H2;
 auto.
generalize (dstep_pop_loop tmp ((r1, r2) :: t) nil); unfold S; simpl app;
 intros H2; apply H2; apply splitNone; assumption.
rewrite stepf_popLoop; auto.
generalize (dstep_pop_loop tmp t (m0 :: l0)); unfold S; simpl; intros H2;
 apply H2; apply splitNone; assumption.
(*cas Pop *)
intro; rewrite stepf_pop; auto.
generalize (dstep_pop tmp t (head_but_last (m :: b))); unfold S; intros H2;
 apply H2; auto.
apply splitNone; assumption.
Qed.
 
Lemma f2ind':
 forall (S1 : State),
 (forall (l : Moves),  (S1 <> (nil, nil, l))) ->  dstep tmp S1 (stepf S1).
intros S1 H; apply f2ind.
trivial.
trivial.
Qed.
 
Lemma ind2f: forall (S1 S2 : State), dstep tmp S1 S2 ->  stepf S1 = S2.
destruct S1 as [[t1 b1] m1]; destruct S2 as [[t2 b2] m2]; intro; simpl.
inversion H.
case (eq_nat_dec r r); intuition.
case (eq_nat_dec s d); intuition.
CaseEq (t0 ++ ((d, r) :: t3)); intros.
destruct t0; destruct t3; discriminate.
rewrite <- H7; destruct m; rewrite splitSome'; auto.
rewrite noRead_noSplit; [rewrite last_app; rewrite last_replace; simpl | idtac];
 auto.
case (eq_nat_dec r0 r0); intuition.
CaseEq (b ++ ((r0, d) :: nil)).
intro; destruct b; discriminate.
intros; destruct t2; (try destruct m0); auto.
rewrite noRead_noSplit; [rewrite last_app; rewrite last_replace; simpl | idtac];
 auto.
case (eq_nat_dec dn s0); intuition.
CaseEq (b ++ ((s0, d0) :: nil)).
intro; destruct b; discriminate.
intros; destruct t2; (try destruct m0); auto.
rewrite noRead_noSplit; auto.
destruct t2; (try destruct m); auto.
Qed.


Lemma appcons_length:
 forall (l1 l2 : Moves) (m : Move),
  length (l1 ++ (m :: l2)) = length (l1 ++ l2) + 1.
induction l1; simpl; intros; [omega | idtac].
rewrite IHl1; omega.
Qed.
 
Definition mesure (S0 : State) : nat :=
   let (p, _) := S0 in let (t, b) := p in 2 * length t + length b.
 
Lemma step_dec0:
 forall (t1 t2 b1 b2 : Moves) (l1 l2 : Moves),
 dstep tmp (t1, b1, l1) (t2, b2, l2) ->
  (2 * length t2 + length b2 < 2 * length t1 + length b1).
intros t1 t2 b1 b2 l1 l2 H; inversion H; simpl; (try omega).
rewrite appcons_length; omega.
cut (length (b ++ ((tmp, d) :: nil)) = length (b ++ ((r0, d) :: nil)));
 (try omega).
induction b; simpl; auto.
rewrite appcons_length.
rewrite appcons_length.
omega.
Qed.
 
Lemma step_dec:
 forall (S1 S2 : State), dstep tmp S1 S2 ->  (mesure S2 < mesure S1).
unfold mesure; destruct S1 as [[t1 b1] l1]; destruct S2 as [[t2 b2] l2].
intro; apply (step_dec0 t1 t2 b1 b2 l1 l2); trivial.
Qed.
 
Lemma stepf_dec0':
 forall (S1 S2 : State),
 (forall (l : Moves),  (S1 <> (nil, nil, l))) ->
  (mesure (stepf S1) < mesure S1).
intros S1 S2 H; apply step_dec.
apply f2ind; trivial.
Qed.
 
Lemma stepf_dec0:
 forall (S1 S2 : State),
 (forall (l : Moves),  (S1 <> (nil, nil, l))) /\ S2 = stepf S1 ->
  (mesure S2 < mesure S1).
intros S1 S2 H; elim H; intros; apply step_dec.
apply f2ind; trivial.
rewrite H1; reflexivity.
Qed.
 
Lemma stepf_dec:
 forall (S1 S2 : State),
 S2 = stepf S1 /\ (forall (l : Moves),  (S1 <> (nil, nil, l))) ->
  ltof _ mesure S2 S1.
unfold ltof.
intros S1 S2 H; elim H; intros; apply step_dec.
apply f2ind; trivial.
rewrite H0; reflexivity.
Qed.
 
Lemma stepf1_dec:
 forall (S1 S2 : State),
 (forall (l : Moves),  (S1 <> (nil, nil, l))) ->
 S2 = stepf S1 ->  ltof _ mesure S2 S1.
unfold ltof.
intros S1 S2 H H'.
apply step_dec.
apply f2ind; trivial.
rewrite H'; reflexivity.
Qed.
 
Lemma disc1:
 forall (a : Move) (l1 l2 l3 l4 : list Move),
  ((a :: l1, l2, l3) <> (nil, nil, l4)).
intros; discriminate.
Qed.
 
Lemma disc2:
 forall (a : Move) (l1 l2 l3 l4 : list Move),
  ((l1, a :: l2, l3) <> (nil, nil, l4)).
intros; discriminate.
Qed.
Hint Resolve disc1 disc2 .
Variables (Value : Set) (v0 : Value).

Lemma sameExec_transitive:
 forall (r1 r2 r3 : State),
 sameExec tmp Value v0 r1 r2 ->
 sameExec tmp Value v0 r2 r3 ->  sameExec tmp Value v0 r1 r3.
intros r1 r2 r3; unfold sameExec.
unfold sameEnv.
unfold exec.
destruct r1; destruct r2; destruct r3.
destruct p; destruct p0.
destruct p1.
intros.
rewrite H; auto.
Qed.
 
Lemma sameExec_reflexive: forall (r : State),  sameExec tmp Value v0 r r.
intros r; unfold sameExec, sameEnv, exec.
destruct r.
destruct p.
auto.
Qed.
 
Lemma In_noWrite_False:
 forall (p : Moves) (s d : Reg), In (s, d) p ->  not (noWrite p d).
intros p s d H H0; (try exact H0).
induction p.
inversion H.
simpl in H0 |-.
destruct a.
elim H.
intro.
inversion H1.
elim H0; intros.
absurd (r0 = d); auto.
intros; apply IHp; auto.
elim H0; intros; auto.
Qed.

Recursive Definition Pmov (State->State)  (ltof _ mesure) 
        (well_founded_ltof _ mesure)  stepf1_dec 
 (forall S1,
 Pmov S1 = match  S1 with ((nil,nil),_) => S1 |_ => Pmov (stepf S1) end).
 
Theorem Fpmov_correct:
 forall (S1 : State), stepInv tmp S1 ->  sameExec tmp Value v0 S1 (Pmov S1).
intros S1; elim S1  using (well_founded_ind (Wf_nat.well_founded_ltof _ mesure)).
clear S1; intros S1; intros Hrec Hinv; rewrite Pmov_equation.
CaseEq S1; intros p m.
CaseEq p; intros t b.
intros.
destruct t.
destruct b.
apply sameExec_reflexive.
rewrite <- H0.
assert (dstep tmp S1 (stepf S1)); (try apply f2ind); auto.
intros; (try assumption).
rewrite H0; auto.
apply sameExec_transitive with (stepf S1).
apply dstepp_sameExec.
apply dstepp_trans with (stepf S1); auto.
apply dstepp_refl.
trivial.
apply Hrec; auto.
unfold ltof; apply step_dec; assumption.
trivial.
apply (dstep_inv tmp S1); assumption.
rewrite <- H0.
assert (dstep tmp S1 (stepf S1)); (try apply f2ind); auto.
intros; (try assumption).
rewrite H0; auto.
apply sameExec_transitive with (stepf S1).
apply dstepp_sameExec.
apply dstepp_trans with (stepf S1); auto.
apply dstepp_refl.
trivial.
apply Hrec; auto.
unfold ltof; apply step_dec; assumption.
trivial.
apply (dstep_inv tmp S1); assumption.
Qed.
 

Definition P_move := fun (p : Moves) => StateDone (Pmov (p, nil, nil)).
 
Definition Sexec := sexec Value v0.
 
Definition Get := get Value v0.

Extraction "pmov4.ml" P_move.
