Require Export List.
Require Export pmov.
Require Export fpmov.
Require Export Bool_nat.


 

Lemma rebuild_l:
 forall (l : Moves) (m : Move),
  m :: l = head_but_last (m :: l) ++ (last (m :: l) :: nil).
induction l.
simpl; auto.
simpl.
intros m; (try assumption).
rewrite (IHl a).
auto.
Qed.
 
Lemma splitNone:
 forall (l : Moves) (r : Reg), split_move l r = None ->  noRead l r.
induction l; intros r; simpl; auto.
destruct a; simpl.
case (eq_nat_dec r0 r); auto.
intros; discriminate.
CaseEq (split_move l r); simpl.
intros p H0; (repeat destruct p).
intros; discriminate.
intros H0 n H1; split; auto.
Qed.
 
Lemma splitSome:
 forall (l t1 t2 : Moves) (s d : Reg),
 split_move l s = Some (t1, d, t2) ->  noRead t1 s.
induction l; simpl.
intros; discriminate.
destruct a.
intros t1 t2 s d; case (eq_nat_dec r s).
intros e H1; inversion H1.
simpl; auto.
CaseEq (split_move l s).
repeat destruct p.
intros H n H0; inversion H0.
simpl.
split; auto.
apply (IHl m0 m s r1); auto.
intros; discriminate.
Qed.
 
Lemma splitSome':
 forall (t1 t2 : Moves) (s d : Reg),
 noRead t1 s ->  split_move (t1 ++ ((s, d) :: t2)) s = Some (t1, d, t2).
induction t1.
intros; simpl.
case (eq_nat_dec s s); auto.
intros n; elim n; auto.
intros t2 s d H; simpl.
destruct a; elim H.
intros; case (eq_nat_dec r s).
intro; absurd (r = s); auto.
intros; rewrite IHt1; auto.
Qed.
 
Lemma unsplit_move:
 forall (l t1 t2 : Moves) (s d : Reg),
 split_move l s = Some (t1, d, t2) ->  l = t1 ++ ((s, d) :: t2).
induction l; simpl.
intros; discriminate.
destruct a.
intros t1 t2 s d; case (eq_nat_dec r s).
intros e H1; inversion H1.
simpl.
rewrite e; auto.
CaseEq (split_move l s).
repeat destruct p.
intros H n H0; inversion H0.
rewrite app_cons.
rewrite (IHl m0 m s r1); auto.
rewrite H3; rewrite H4; auto.
intros; discriminate.
Qed.
 
Lemma cons_replace:
 forall (a : Move) (l : Moves),
 l <> nil ->  replace_last_s (a :: l) = a :: replace_last_s l.
intros; simpl.
CaseEq l.
intro; contradiction.
intros m l0 H0; auto.
Qed.
 
Lemma last_replace:
 forall (l : Moves) (s d : Reg),
  replace_last_s (l ++ ((s, d) :: nil)) = l ++ ((tmp, d) :: nil).
induction l.
simpl; auto.
intros.
rewrite <- app_comm_cons.
rewrite <- app_comm_cons.
rewrite cons_replace.
rewrite IHl; auto.
case l; auto.
simpl.
discriminate.
intros m l0 H; (try exact H).
discriminate H.
Qed.
 
Lemma last_app: forall (l : Moves) (m : Move),  last (l ++ (m :: nil)) = m.
induction l.
simpl; auto.
simpl.
intros m; CaseEq (l ++ (m :: nil)).
intro; destruct l; discriminate.
intros m0 l0 H; (rewrite <- H; (try assumption)).
apply IHl.
Qed.
 
Lemma noRead_noSplit:
 forall (l : Moves) (r : Reg), noRead l r ->  split_move l r = None.
intros l r H; (try assumption).
induction l.
simpl; auto.
simpl.
destruct a; auto.
elim H.
intros.
case (eq_nat_dec r0 r).
intro; absurd (r = r0); auto.
intro; rewrite IHl; auto.
Qed.
 
Lemma last_cons:
 forall (l : Moves) (m m0 : Move),  last (m0 :: (m :: l)) = last (m :: l).
intros; simpl; auto.
Qed.
 
Lemma stepf_popLoop:
 forall t b l : Moves,
 forall s d r0 : Reg,
 split_move t d = None ->
  stepf (t, (s, d) :: (b ++ ((d, r0) :: nil)), l) =
  (t, b ++ ((tmp, r0) :: nil), (s, d) :: ((d, tmp) :: l)).
intros; simpl; rewrite H; CaseEq (b ++ ((d, r0) :: nil)); intros.
destruct b; discriminate.
rewrite <- H0; rewrite last_app; simpl; rewrite last_replace.
case (eq_nat_dec d d); intro; intuition.
destruct t; (try destruct m0); simpl; auto.
Qed.
 
Lemma stepf_pop:
 forall t b l : Moves,
 forall s d r r0 : Reg,
 split_move t d = None ->
 d <> r ->
  stepf (t, (s, d) :: (b ++ ((r, r0) :: nil)), l) =
  (t, b ++ ((r, r0) :: nil), (s, d) :: l).
intros; simpl; rewrite H; CaseEq (b ++ ((r, r0) :: nil)); intros.
destruct b; discriminate.
rewrite <- H1; rewrite last_app; simpl.
case (eq_nat_dec d r); intro; intuition.
destruct t; (try destruct m0); simpl; auto.
Qed.
 