Require Export TheoryList.
Require Export Bool.
Require Export Arith.
Require Import Peano_dec.
Require Export EqNat.
 
Section pmov.
 
Axiom Skip : False.
 
Ltac Damn := cut False; [intros Fa; inversion Fa | apply Skip].
 
Ltac
Switch PRED RES :=
cut (exists b : bool , b = PRED );
 [intros [BBBBB RES]; generalize RES; case PRED; intros FFFFF;
   rewrite FFFFF in RES; clear FFFFF BBBBB | exists PRED]; auto.
 
Ltac
Rename NAME TYPE EXPR EQ :=
cut (exists NAME : TYPE , NAME = EXPR );
 [intros [NAME EQ]; rewrite <- EQ | exists EXPR]; auto.
 
Ltac Use EQ := cut EQ; [intros BBBBB; rewrite BBBBB; clear BBBBB | eauto].
 
Ltac ACut PPPPP HHHHH := cut PPPPP; [intros HHHHH | auto].
 
Ltac ACut2 PPPPP CAR CDR := cut PPPPP; [intros [CAR CDR] | auto].
 
Ltac ECut PPPPP HHHHH := cut PPPPP; [intros HHHHH | eauto].
 
Ltac TestEqNat X Y Z := Switch (beq_nat X Y) Z; auto; Use (X = Y); auto.
(* * External lemmas
   *)
 
Lemma beq_nat_eq: forall (x y : nat), true = beq_nat x y ->  x = y.
Proof.
intros n1; induction n1 as [|n1 Hrecn1]; intros n2;
 induction n2 as [|n2 Hrecn2]; auto; simpl; intros H; inversion H.
Qed.
Hint Resolve beq_nat_eq .
 
Lemma eq_is_eq: forall (x : nat),  beq_nat x x = true.
Proof.
intros x; induction x as [|x Hrecx]; auto.
Qed.
Hint Resolve eq_is_eq .
 
Lemma neq_is_neq: forall (x y : nat), x <> y ->  beq_nat x y = false.
Proof.
unfold not; intros.
Switch (beq_nat x y) ipattern:EQ; auto.
generalize H; Use (x = y); intros.
cut False; auto.
intros; inversion H1.
Qed.
Hint Resolve neq_is_neq .
 
Lemma app_nil: forall (A : Set) (l : list A),  l ++ nil = l.
Proof.
intros A l; induction l as [|a l Hrecl]; auto; simpl; rewrite Hrecl; auto.
Qed.
 
Lemma app_cons:
 forall (A : Set) (l1 l2 : list A) (a : A),  (a :: l1) ++ l2 = a :: (l1 ++ l2).
Proof.
auto.
Qed.
 
Lemma app_app:
 forall (A : Set) (l1 l2 l3 : list A),  l1 ++ (l2 ++ l3) = (l1 ++ l2) ++ l3.
Proof.
intros A l1; induction l1 as [|a l1 Hrecl1]; simpl; auto; intros;
 rewrite Hrecl1; auto.
Qed.
 
Lemma app_rewrite:
 forall (A : Set) (l : list A) (x : A),
  (exists y : A , exists r : list A , l ++ (x :: nil) = y :: r  ).
Proof.
intros A l x; induction l as [|a l Hrecl].
exists x; exists (nil (A:=A)); auto.
inversion Hrecl; inversion H.
exists a; exists (l ++ (x :: nil)); auto.
Qed.
 
Lemma app_rewrite2:
 forall (A : Set) (l l2 : list A) (x : A),
  (exists y : A , exists r : list A , l ++ (x :: l2) = y :: r  ).
Proof.
intros A l l2 x; induction l as [|a l Hrecl].
exists x; exists l2; auto.
inversion Hrecl; inversion H.
exists a; exists (l ++ (x :: l2)); auto.
Qed.
 
Lemma app_rewriter:
 forall (A : Set) (l : list A) (x : A),
  (exists y : A , exists r : list A , x :: l = r ++ (y :: nil)  ).
Proof.
intros A l x; induction l as [|a l Hrecl].
exists x; exists (nil (A:=A)); auto.
inversion Hrecl; inversion H.
generalize H0; case x1; simpl; intros; inversion H1.
exists a; exists (x0 :: nil); simpl; auto.
exists x0; exists (a0 :: (a :: l0)); simpl; auto.
Qed.
Hint Resolve app_rewriter .
(* * Data definitions
   *)
 
Definition Reg := nat.
Variable tmp : Reg.
 
Definition Move := (Reg * Reg)%type.
 
Definition Moves := list Move.
 
Definition State := ((Moves * Moves) * Moves)%type.
 
Definition StateToMove (r : State) : Moves :=
   match r with ((t, b), l) => t end.
 
Definition StateBeing (r : State) : Moves :=
   match r with ((t, b), l) => b end.
 
Definition StateDone (r : State) : Moves :=
   match r with ((t, b), l) => l end.
 
Definition S (t b l : Moves) : State := (t, b, l).
 
Ltac
Rewrite_app_cons EXPR M B EQ :=
cut (exists M : Move , exists B : Moves , EXPR = M :: B  );
 [intros [M [B EQ]]; rewrite EQ | apply (app_rewrite Move)].
 
Ltac
Rewrite_app_cons2 EXPR M B EQ :=
cut (exists M : Move , exists B : Moves , EXPR = M :: B  );
 [intros [M [B EQ]]; rewrite EQ | apply (app_rewrite2 Move)].
 
Ltac
Rewrite_cons_app EXPR M B EQ :=
cut (exists M : Move , exists B : Moves , EXPR = B ++ (M :: nil)  );
 [intros [M [B EQ]]; rewrite EQ | apply (app_rewriter Move)].
(* * A predicate showing that a list of moves doesn't read a specific register
   *)
 
Fixpoint noRead (p : Moves) (r : Reg) {struct p} : Prop :=
 match p with
   nil => True
  | (s, d) :: l => s <> r /\ noRead l r
 end.
 
Lemma noRead_app:
 forall (l1 l2 : Moves) (r : Reg),
 noRead l1 r -> noRead l2 r ->  noRead (l1 ++ l2) r.
Proof.
intros; induction l1 as [|a l1 Hrecl1]; simpl; auto.
generalize H; case a; simpl.
intros r0 _ [NEQ NR]; split; auto.
Qed.
(* * A step of the algorithm
   *)
 
Inductive step : State -> State ->  Prop :=
  step_nop:
    forall (r : Reg) (t1 t2 l : Moves),
     step (S (t1 ++ ((r, r) :: t2)) nil l) (S (t1 ++ t2) nil l)
 | step_start:
     forall (t1 t2 l : Moves) (m : Move),
      step (S (t1 ++ (m :: t2)) nil l) (S (t1 ++ t2) (m :: nil) l)
 | step_push:
     forall (t1 t2 b l : Moves) (s d r : Reg),
      step
       (S (t1 ++ ((d, r) :: t2)) ((s, d) :: b) l)
       (S (t1 ++ t2) ((d, r) :: ((s, d) :: b)) l)
 | step_loop:
     forall (t b l : Moves) (s d r0 r0ounon : Reg),
      step
       (S t ((s, r0ounon) :: (b ++ ((r0, d) :: nil))) l)
       (S t ((s, r0ounon) :: (b ++ ((tmp, d) :: nil))) ((r0, tmp) :: l))
 | step_pop:
     forall (t b l : Moves) (s0 d0 sn dn : Reg),
     noRead t dn ->
     dn <> s0 ->
      step
       (S t ((sn, dn) :: (b ++ ((s0, d0) :: nil))) l)
       (S t (b ++ ((s0, d0) :: nil)) ((sn, dn) :: l))
 | step_last:
     forall (t l : Moves) (s d : Reg),
     noRead t d ->  step (S t ((s, d) :: nil) l) (S t nil ((s, d) :: l)) .
Hint Resolve step_nop step_start step_push step_loop step_pop step_last .
(* *
   *)
 
Fixpoint path (l : Moves) : Prop :=
 match l with
   nil => True
  | (r8, r9) :: l =>
      match l with
        nil => True
       | (r7, rr8) :: l2 => r8 = rr8 /\ path l
      end
 end.
 
Lemma path_pop: forall (m : Move) (l : Moves), path (m :: l) ->  path l.
Proof.
simpl; intros m l; case m; case l; auto.
intros m0; case m0; intros; inversion H; auto.
Qed.
(* *)
 
Fixpoint noWrite (p : Moves) (r : Reg) {struct p} : Prop :=
 match p with
   nil => True
  | (s, d) :: l => d <> r /\ noWrite l r
 end.
 
Lemma noWrite_pop:
 forall (p1 p2 : Moves) (m : Move) (r : Reg),
 noWrite (p1 ++ (m :: p2)) r ->  noWrite (p1 ++ p2) r.
Proof.
intros; induction p1 as [|a p1 Hrecp1].
generalize H; simpl; case m; intros; inversion H0; auto.
generalize H; rewrite app_cons; rewrite app_cons.
simpl; case a; intros.
inversion H0; split; auto.
Qed.
 
Lemma noWrite_in:
 forall (p1 p2 : Moves) (r0 r1 r2 : Reg),
 noWrite (p1 ++ ((r1, r2) :: p2)) r0 ->  (r0 <> r2).
Proof.
intros; induction p1 as [|a p1 Hrecp1]; simpl; auto.
generalize H; simpl; intros; inversion H0; auto.
generalize H; rewrite app_cons; simpl; case a; intros.
apply Hrecp1; inversion H0; auto.
Qed.
 
Lemma noWrite_swap:
 forall (p : Moves) (m1 m2 : Move) (r : Reg),
 noWrite (m1 :: (m2 :: p)) r ->  noWrite (m2 :: (m1 :: p)) r.
Proof.
intros p m1 m2 r; simpl; case m1; case m2.
intros; inversion H; inversion H1; split; auto.
Qed.
 
Lemma noWrite_movFront:
 forall (p1 p2 : Moves) (m : Move) (r0 : Reg),
 noWrite (p1 ++ (m :: p2)) r0 ->  noWrite (m :: (p1 ++ p2)) r0.
Proof.
intros p1 p2 m r0; induction p1 as [|a p1 Hrecp1]; auto.
case a; intros r1 r2; rewrite app_cons; rewrite app_cons.
intros; apply noWrite_swap; rewrite <- app_cons.
simpl in H |-; inversion H; unfold noWrite; fold noWrite; auto.
Qed.
 
Lemma noWrite_insert:
 forall (p1 p2 : Moves) (m : Move) (r0 : Reg),
 noWrite (m :: (p1 ++ p2)) r0 ->  noWrite (p1 ++ (m :: p2)) r0.
Proof.
intros p1 p2 m r0; induction p1 as [|a p1 Hrecp1].
simpl; auto.
rewrite app_cons; rewrite app_cons.
intros; ACut (noWrite (a :: m :: p1 ++ p2) r0) ipattern:NW.
 ECut (noWrite (p1 ++ m :: p2) r0) ipattern:NW2.
generalize NW; unfold noWrite; fold noWrite.
case a; split; auto; inversion NW0; auto.
apply Hrecp1; generalize NW; simpl; case a; case m; intros.
inversion NW0; auto.
apply noWrite_swap; auto.
Qed.
 
Lemma noWrite_tmpLast:
 forall (t : Moves) (r s d : Reg),
 noWrite (t ++ ((s, d) :: nil)) r ->  noWrite (t ++ ((tmp, d) :: nil)) r.
Proof.
intros; induction t as [|a t Hrect].
simpl; auto.
generalize H; simpl; case a; intros; inversion H0; split; auto.
Qed.
(* *)
 
Fixpoint simpleDest (p : Moves) : Prop :=
 match p with
   nil => True
  | (s, d) :: l => noWrite l d /\ simpleDest l
 end.
 
Lemma simpleDest_Pop:
 forall (m : Move) (l1 l2 : Moves),
 simpleDest (l1 ++ (m :: l2)) ->  simpleDest (l1 ++ l2).
Proof.
intros; induction l1 as [|a l1 Hrecl1].
generalize H; simpl; case m; intros; inversion H0; auto.
generalize H; rewrite app_cons; rewrite app_cons.
simpl; case a; intros; inversion H0; split; auto.
apply (noWrite_pop l1 l2 m); auto.
Qed.
 
Lemma simpleDest_pop:
 forall (m : Move) (l : Moves), simpleDest (m :: l) ->  simpleDest l.
Proof.
intros m l; simpl; case m; intros _ r [X Y]; auto.
Qed.
 
Lemma simpleDest_right:
 forall (l1 l2 : Moves), simpleDest (l1 ++ l2) ->  simpleDest l2.
Proof.
intros l1; induction l1 as [|a l1 Hrecl1]; auto.
intros l2; rewrite app_cons; intros; apply Hrecl1.
apply (simpleDest_pop a); auto.
Qed.
 
Lemma simpleDest_swap:
 forall (m1 m2 : Move) (l : Moves),
 simpleDest (m1 :: (m2 :: l)) ->  simpleDest (m2 :: (m1 :: l)).
Proof.
intros m1 m2 l; simpl; case m1; case m2.
intros _ r0 _ r2 [[X Y] [Z T]]; auto.
Qed.
 
Lemma simpleDest_pop2:
 forall (m1 m2 : Move) (l : Moves),
 simpleDest (m1 :: (m2 :: l)) ->  simpleDest (m1 :: l).
Proof.
intros; apply (simpleDest_pop m2); apply simpleDest_swap; auto.
Qed.
 
Lemma simpleDest_movFront:
 forall (p1 p2 : Moves) (m : Move),
 simpleDest (p1 ++ (m :: p2)) ->  simpleDest (m :: (p1 ++ p2)).
Proof.
intros p1 p2 m; induction p1 as [|a p1 Hrecp1].
simpl; auto.
rewrite app_cons; rewrite app_cons.
case a; intros; simpl in H |-; inversion H.
ACut (simpleDest (m :: p1 ++ p2)) ipattern:SD.
ACut (noWrite (m :: p1 ++ p2) r0) ipattern:NW.
apply simpleDest_swap; simpl; auto.
apply noWrite_movFront; auto.
Qed.
 
Lemma simpleDest_insert:
 forall (p1 p2 : Moves) (m : Move),
 simpleDest (m :: (p1 ++ p2)) ->  simpleDest (p1 ++ (m :: p2)).
Proof.
intros p1 p2 m; induction p1 as [|a p1 Hrecp1].
simpl; auto.
rewrite app_cons; intros.
ACut (simpleDest (a :: m :: p1 ++ p2)) ipattern:SD.
 ACut (simpleDest (p1 ++ m :: p2)) ipattern:SD2.
rewrite app_cons.
generalize SD; unfold simpleDest; fold simpleDest.
case a; split; auto; inversion SD0; auto.
apply noWrite_insert; auto.
apply Hrecp1; apply (simpleDest_pop a); auto.
apply simpleDest_swap; auto.
Qed.
 
Lemma simpleDest_movBack:
 forall (p1 p2 : Moves) (m : Move),
 simpleDest (p1 ++ (m :: p2)) ->  simpleDest ((p1 ++ p2) ++ (m :: nil)).
Proof.
intros.
apply (simpleDest_insert (p1 ++ p2) nil m).
rewrite app_nil; apply simpleDest_movFront; auto.
Qed.
 
Lemma simpleDest_swap_app:
 forall (t1 t2 t3 : Moves) (m : Move),
 simpleDest (t1 ++ (m :: (t2 ++ t3))) ->  simpleDest ((t1 ++ t2) ++ (m :: t3)).
Proof.
intros.
apply (simpleDest_insert (t1 ++ t2) t3 m).
rewrite <- app_app.
apply simpleDest_movFront; auto.
Qed.
 
Lemma simpleDest_tmpLast:
 forall (t : Moves) (s d : Reg),
 simpleDest (t ++ ((s, d) :: nil)) ->  simpleDest (t ++ ((tmp, d) :: nil)).
Proof.
intros t s d; induction t as [|a t Hrect].
simpl; auto.
rewrite app_cons; rewrite app_cons.
simpl; case a; intros; inversion H; split; auto.
apply (noWrite_tmpLast t r0 s); auto.
Qed.
(* *)
 
Fixpoint noTmp (b : Moves) : Prop :=
 match b with
   nil => True
  | (s, d) :: l => s <> tmp /\ (d <> tmp /\ noTmp l)
 end.
 
Fixpoint noTmpLast (b : Moves) : Prop :=
 match b with
   nil => True
  | (s, d) :: nil => d <> tmp
  | (s, d) :: l => s <> tmp /\ (d <> tmp /\ noTmpLast l)
 end.
 
Lemma noTmp_app:
 forall (l1 l2 : Moves) (m : Move),
 noTmp l1 -> noTmpLast (m :: l2) ->  noTmpLast (l1 ++ (m :: l2)).
Proof.
intros.
induction l1 as [|a l1 Hrecl1].
simpl; auto.
rewrite app_cons.
generalize H.
simpl.
case a.
Rewrite_app_cons2 (l1 ++ m :: l2) ipattern:m' ipattern:t' ipattern:EQ.
intros r r0 [NEQ [NEQ2 NTL]].
split; auto.
split; auto.
rewrite <- EQ; auto.
Qed.
 
Lemma noTmpLast_popBack:
 forall (t : Moves) (m : Move), noTmpLast (t ++ (m :: nil)) ->  noTmp t.
Proof.
intros.
induction t as [|a t Hrect].
simpl; auto.
generalize H.
rewrite app_cons.
simpl.
case a.
Rewrite_app_cons (t ++ m :: nil) ipattern:m' ipattern:t' ipattern:EQ.
intros r r0 [NEQ [NEQ2 NTL]].
split; auto.
split; auto.
apply Hrect.
rewrite EQ; auto.
Qed.
(* *)
 
Definition stepInv (r : State) : Prop :=
   path (StateBeing r) /\
   (simpleDest (StateToMove r ++ StateBeing r) /\
    (noTmp (StateToMove r) /\ noTmpLast (StateBeing r))).
(* *
   
   *)
Variable Value : Set.
Variable null : Value.
 
Definition Env := list (Reg * Value).
 
Fixpoint get (e : Env) (r : Reg) {struct e} : Value :=
 match e with
   nil => null
  | (r', v) :: l => if beq_nat r r' then v else get l r
 end.
(* *)
 
Definition update (e : Env) (r : Reg) (v : Value) : Env := (r, v) :: e.
(* *)
 
Fixpoint sexec (p : Moves) (e : Env) {struct p} : Env :=
 match p with
   nil => e
  | (s, d) :: l => let e' := sexec l e in
                     update e' d (get e' s)
 end.
(* *)
 
Fixpoint pexec (p : Moves) (e : Env) {struct p} : Env :=
 match p with
   nil => e
  | (s, d) :: l => update (pexec l e) d (get e s)
 end.
 
Lemma pexec_swap:
 forall (m1 m2 : Move) (t : Moves),
 simpleDest (m1 :: (m2 :: t)) ->
 forall (e : Env) (r : Reg),
  get (pexec (m1 :: (m2 :: t)) e) r = get (pexec (m2 :: (m1 :: t)) e) r.
Proof.
intros.
simpl.
generalize H; case m1; case m2; simpl; intros.
generalize H0; intros [[NEQ NW] R].
Switch (beq_nat r r3) ipattern:EQ1; auto.
Switch (beq_nat r r1) ipattern:EQ2; auto.
generalize EQ2; Use (r = r3); auto; intros EQ3.
generalize NEQ; Use (r3 = r1).
unfold not.
intros BAD.
cut False; auto.
intros XX; inversion XX.
Qed.
 
Lemma pexec_add:
 forall (t1 t2 : Moves) (r : Reg) (e : Env),
 get (pexec t1 e) r = get (pexec t2 e) r ->
 forall (a : Move),  get (pexec (a :: t1) e) r = get (pexec (a :: t2) e) r.
Proof.
intros.
case a.
simpl.
intros.
case (beq_nat r r1); auto.
Qed.
 
Lemma pexec_movBack:
 forall (t1 t2 : Moves) (m : Move),
 simpleDest (m :: (t1 ++ t2)) ->
 forall (e : Env) (r : Reg),
  get (pexec (m :: (t1 ++ t2)) e) r = get (pexec (t1 ++ (m :: t2)) e) r.
Proof.
intros t1 t2 m; induction t1 as [|a t1 Hrect1].
simpl; auto.
rewrite app_cons.
intros; rewrite pexec_swap; auto; rewrite app_cons; auto.
apply pexec_add.
apply Hrect1.
apply (simpleDest_pop2 m a); auto.
Qed.
 
Lemma pexec_movFront:
 forall (t1 t2 : Moves) (m : Move),
 simpleDest (t1 ++ (m :: t2)) ->
 forall (e : Env) (r : Reg),
  get (pexec (t1 ++ (m :: t2)) e) r = get (pexec (m :: (t1 ++ t2)) e) r.
Proof.
intros; rewrite <- pexec_movBack; eauto.
apply simpleDest_movFront; auto.
Qed.
 
Lemma pexec_mov:
 forall (t1 t2 t3 : Moves) (m : Move),
 simpleDest ((t1 ++ (m :: t2)) ++ t3) ->
 forall (e : Env) (r : Reg),
  get (pexec ((t1 ++ (m :: t2)) ++ t3) e) r =
  get (pexec ((t1 ++ t2) ++ (m :: t3)) e) r.
Proof.
intros t1 t2 t3 m.
rewrite <- app_app.
rewrite app_cons.
intros.
rewrite pexec_movFront; auto.
cut (simpleDest (m :: (t1 ++ (t2 ++ t3)))).
rewrite app_app.
rewrite <- pexec_movFront; auto.
apply simpleDest_swap_app; auto.
apply simpleDest_movFront; auto.
Qed.
 
Lemma pexec_nop:
 forall (t : Moves) (r : Reg),
 simpleDest ((r, r) :: t) ->
 forall (e : Env) (x : Reg),  get (pexec ((r, r) :: t) e) x = get (pexec t e) x.
Proof.
intros.
simpl.
Switch (beq_nat x r) ipattern:EQ; auto.
Use (x = r).
induction t as [|a t Hrect].
simpl; auto.
generalize H; case a; simpl.
intros r0 r1 [[NEQ NW] R].
Use (beq_nat r r1 = false); auto.
apply Hrect.
apply (simpleDest_pop2 (r, r) a); auto.
Qed.
(*** USEFULL ***)
 
Lemma get_update:
 forall (e : Env) (r1 r2 : Reg) (v : Value),
  get (update e r1 v) r2 = (if beq_nat r2 r1 then v else get e r2).
Proof.
simpl; intros; case e; simpl; auto.
Qed.
 
Lemma get_noWrite:
 forall (t : Moves) (d : Reg),
 noWrite t d ->
 forall (e : Env) (v : Value),  v = get (pexec t (update e d v)) d.
Proof.
intros.
induction t as [|a t Hrect].
simpl.
Use (beq_nat d d = true); auto.
generalize H; case a; simpl; intros r r0 [NEQ R].
Use (beq_nat d r0 = false); auto.
Qed.
 
Lemma pexec_push:
 forall (t l : Moves) (s d : Reg),
 noRead t d ->
 simpleDest ((s, d) :: t) ->
 forall (e : Env) (r : Reg),
  get (pexec ((s, d) :: t) (sexec l e)) r =
  get (pexec t (sexec ((s, d) :: l) e)) r.
Proof.
intros.
simpl.
Switch (beq_nat r d) ipattern:EQ.
Use (r = d).
apply get_noWrite.
inversion H0; auto.
induction t as [|a t Hrect].
simpl.
rewrite <- EQ; auto.
generalize H H0; case a; simpl.
intros r0 r1 [NEQ NR] [[NEQ2 NW] R].
Switch (beq_nat r r1) ipattern:EQ2.
 Switch (beq_nat r0 d) ipattern:EQ3; auto.
generalize NEQ; Use (r0 = d).
unfold not.
intros BAD.
cut False; auto.
intros XX; inversion XX.
apply Hrect; auto.
apply (simpleDest_pop2 (s, d) a); auto.
Qed.
(*Lemma nowrite_get:
   (noRead t d)
   (simpleDest (cons (s,d) t))
   (get e s)=(get (pexec t (update e d (get e s))) d)
   * Step Correctness
   *)
 
Definition exec (s : State) (e : Env) :=
   pexec (StateToMove s ++ StateBeing s) (sexec (StateDone s) e).
 
Definition sameEnv (e1 e2 : Env) :=
   forall (r : Reg), r <> tmp ->  get e1 r = get e2 r.
 
Definition sameExec (s1 s2 : State) :=
   forall (e : Env),  sameEnv (exec s1 e) (exec s2 e).
(* 
   * Usefull lemmas*)
 
Lemma update_tmp:
 forall (t : Moves) (r : Reg),
 noTmp t ->
 r <> tmp ->
 forall (e : Env) (v : Value),
  get (pexec t e) r = get (pexec t (update e tmp v)) r.
Proof.
intros t; induction t as [|a t Hrect].
intros; simpl.
Use (beq_nat r tmp = false); auto.
intros.
simpl.
generalize H; case a; simpl.
intros r0 r1 [N1 [N2 RT]].
case (beq_nat r r1); auto.
Use (beq_nat r0 tmp = false); auto.
Qed.
 
Lemma step_inv_loop_aux:
 forall (t l : Moves) (s d : Reg),
 simpleDest (t ++ ((s, d) :: nil)) ->
 noTmp t ->
 forall (e : Env) (r : Reg),
 r <> tmp ->
  get (pexec (t ++ ((s, d) :: nil)) (sexec l e)) r =
  get (pexec (t ++ ((tmp, d) :: nil)) (sexec ((s, tmp) :: l) e)) r.
Proof.
intros.
rewrite pexec_movFront; auto.
rewrite pexec_movFront; [rewrite app_nil | apply (simpleDest_tmpLast t s)]; auto.
simpl.
Use (beq_nat tmp tmp = true); auto.
case (beq_nat r d); auto.
apply update_tmp; auto.
Qed.
 
Lemma step_inv_loop:
 forall (t l : Moves) (s d : Reg),
 simpleDest (t ++ ((s, d) :: nil)) ->
 noTmpLast (t ++ ((s, d) :: nil)) ->
 forall (e : Env) (r : Reg),
 r <> tmp ->
  get (pexec (t ++ ((s, d) :: nil)) (sexec l e)) r =
  get (pexec (t ++ ((tmp, d) :: nil)) (sexec ((s, tmp) :: l) e)) r.
Proof.
intros.
apply step_inv_loop_aux; auto.
apply (noTmpLast_popBack t (s, d)); auto.
Qed.
 
Lemma noRead_by_path:
 forall (b t : Moves) (r0 r1 r7 r8 : Reg),
 simpleDest ((r7, r8) :: (b ++ ((r0, r1) :: nil))) ->
 path (b ++ ((r0, r1) :: nil)) -> r8 <> r0 ->  noRead b r8.
Proof.
intros; induction b as [|a b Hrecb]; simpl; auto.
generalize H H0; rewrite app_cons; case a; intros; split.
 generalize H2 H3;
  Rewrite_app_cons (b ++ (r0, r1) :: nil) ipattern:m ipattern:b' ipattern:EQ.
simpl; case m; intros r3 r4 [[_ [NEQ _]] _] [EQ2 _]; rewrite EQ2; auto.
apply Hrecb; eauto.
apply (simpleDest_pop (r, r2)).
apply simpleDest_swap; auto.
apply (path_pop (r, r2)); auto.
Qed.
 
Lemma step_sameExec:
 forall (r1 r2 : State), step r1 r2 -> stepInv r1 ->  sameExec r1 r2.
Proof.
intros r1 r2 STEP; inversion_clear STEP;
 unfold S, stepInv, sameExec, sameEnv, exec, StateToMove, StateBeing, StateDone;
 intros [P [SD [TT TB]]]; intros.
(* Nop *)
rewrite app_nil; rewrite app_nil; rewrite app_nil in SD.
rewrite pexec_movFront; auto.
rewrite pexec_nop; auto.
apply simpleDest_movFront; auto.
(* Start *)
rewrite app_nil; rewrite app_nil in SD.
rewrite pexec_movFront; auto.
rewrite pexec_movFront; auto.
rewrite app_nil; auto.
apply simpleDest_movBack; auto.
(* Push *)
apply pexec_mov; auto.
(* Loop *)
rewrite <- app_cons; rewrite app_app.
rewrite <- app_cons; rewrite app_app.
apply step_inv_loop; auto.
rewrite <- app_app; rewrite app_cons; auto.
rewrite <- app_app; rewrite app_cons; auto.
apply noTmp_app; auto.
(* Pop *)
rewrite pexec_movFront; auto.
apply pexec_push; auto.
apply noRead_app; auto.
apply noRead_app.
apply (noRead_by_path b b s0 d0 sn dn); auto.
apply (simpleDest_right t); auto.
apply (path_pop (sn, dn)); auto.
simpl; split; auto.
apply simpleDest_movFront; auto.
(* Last *)
rewrite pexec_movFront; auto.
apply pexec_push; auto.
rewrite app_nil; auto.
apply simpleDest_movFront; auto.
Qed.
(* STEP INV *)
 
Lemma path_tmpLast:
 forall (s d : Reg) (l : Moves),
 path (l ++ ((s, d) :: nil)) ->  path (l ++ ((tmp, d) :: nil)).
Proof.
intros; induction l as [|a l Hrecl].
simpl; auto.
generalize H; rewrite app_cons; rewrite app_cons; simpl.
case a; generalize Hrecl; case l.
simpl; auto.
intros m l0 REC r _; generalize REC; rewrite app_cons; rewrite app_cons.
case m; intros; inversion H0; split; auto.
Qed.
 
Lemma step_inv_path:
 forall (r1 r2 : State), step r1 r2 -> stepInv r1 ->  path (StateBeing r2).
Proof.
intros r1 r2 STEP; inversion_clear STEP; unfold stepInv;
 unfold S, stepInv, sameExec, sameEnv, exec, StateToMove, StateBeing, StateDone;
 intros [P [SD [TT TB]]].
(* Nop *)
simpl; auto.
(* Start *)
simpl; case m; auto.
(* Push *)
generalize P; simpl; intros; split; auto.
(* Loop *)
generalize P; rewrite <- app_cons; rewrite <- app_cons.
apply (path_tmpLast r0).
(* Pop *)
generalize P; apply path_pop.
(* Last *)
simpl; auto.
Qed.
(* *)
 
Lemma step_inv_simpleDest:
 forall (r1 r2 : State),
 step r1 r2 -> stepInv r1 ->  simpleDest (StateToMove r2 ++ StateBeing r2).
Proof.
intros r1 r2 STEP; inversion_clear STEP; unfold stepInv;
 unfold S, stepInv, sameExec, sameEnv, exec, StateToMove, StateBeing, StateDone;
 intros [P [SD [TT TB]]].
(* Nop *)
rewrite app_nil in SD; rewrite app_nil; apply (simpleDest_Pop (r, r)); auto.
(* Start *)
rewrite app_nil in SD; apply simpleDest_movBack; auto.
(* Push *)
apply simpleDest_insert; rewrite <- app_app; apply simpleDest_movFront.
rewrite <- app_cons; rewrite app_app; auto.
(* Loop *)
generalize SD; rewrite <- app_cons; rewrite <- app_cons.
rewrite app_app; rewrite app_app; apply simpleDest_tmpLast.
(* Pop *)
generalize SD; apply simpleDest_Pop.
(* Last *)
generalize SD; apply simpleDest_Pop.
Qed.
(* *)
 
Lemma noTmp_pop:
 forall (m : Move) (l1 l2 : Moves), noTmp (l1 ++ (m :: l2)) ->  noTmp (l1 ++ l2).
Proof.
intros; induction l1 as [|a l1 Hrecl1]; generalize H.
simpl; case m; intros; inversion H0; inversion H2; auto.
rewrite app_cons; rewrite app_cons; simpl; case a.
intros; inversion H0; inversion H2; auto.
Qed.
 
Lemma step_inv_noTmp:
 forall (r1 r2 : State), step r1 r2 -> stepInv r1 ->  noTmp (StateToMove r2).
Proof.
intros r1 r2 STEP; inversion_clear STEP; unfold stepInv;
 unfold S, stepInv, sameExec, sameEnv, exec, StateToMove, StateBeing, StateDone;
 intros [P [SD [TT TB]]]; generalize TT.
(* Nop *)
apply noTmp_pop.
(* Start *)
apply noTmp_pop.
(* Push *)
apply noTmp_pop.
(* Loop *)
auto.
(* Pop *)
auto.
(* Last *)
auto.
Qed.
(* *)
 
Lemma noTmp_noTmpLast: forall (l : Moves), noTmp l ->  noTmpLast l.
Proof.
intros; induction l as [|a l Hrecl].
simpl; auto.
generalize H; simpl; case a; generalize Hrecl.
case l.
intros; inversion H0; inversion H2; auto.
intros; inversion H0; inversion H2; auto.
Qed.
 
Lemma noTmpLast_pop:
 forall (m : Move) (l : Moves), noTmpLast (m :: l) ->  noTmpLast l.
Proof.
intros m l; simpl; case m; case l.
simpl; auto.
intros; inversion H; inversion H1; auto.
Qed.
 
Lemma noTmpLast_Pop:
 forall (m : Move) (l1 l2 : Moves),
 noTmpLast (l1 ++ (m :: l2)) ->  noTmpLast (l1 ++ l2).
Proof.
intros; induction l1 as [|a l1 Hrecl1]; generalize H.
simpl; case m; case l2.
simpl; auto.
intros; inversion H0; inversion H2; auto.
rewrite app_cons; rewrite app_cons; simpl; case a.
generalize Hrecl1; case l1.
simpl; case m; case l2; intros; inversion H0; inversion H2; auto.
intros m0 l R r r0; rewrite app_cons; rewrite app_cons.
intros; inversion H0; inversion H2; auto.
Qed.
 
Lemma noTmpLast_push:
 forall (m : Move) (t1 t2 t3 : Moves),
 noTmp (t1 ++ (m :: t2)) -> noTmpLast t3 ->  noTmpLast (m :: t3).
Proof.
intros; induction t1 as [|a t1 Hrect1]; generalize H.
simpl; case m; intros r r0 [N1 [N2 NT]]; generalize H0; case t3; auto.
rewrite app_cons; intros; apply Hrect1.
generalize H1.
simpl; case m; case a; intros; inversion H2; inversion H4; auto.
Qed.
 
Lemma noTmpLast_tmpLast:
 forall (s d : Reg) (l : Moves),
 noTmpLast (l ++ ((s, d) :: nil)) ->  noTmpLast (l ++ ((tmp, d) :: nil)).
Proof.
intros; induction l as [|a l Hrecl].
simpl; auto.
generalize H; rewrite app_cons; rewrite app_cons; simpl.
case a; generalize Hrecl; case l.
simpl; auto.
intros m l0 REC r r0; generalize REC; rewrite app_cons; rewrite app_cons.
case m; intros; inversion H0; inversion H2; split; auto.
Qed.
 
Lemma step_inv_noTmpLast:
 forall (r1 r2 : State), step r1 r2 -> stepInv r1 ->  noTmpLast (StateBeing r2).
Proof.
intros r1 r2 STEP; inversion_clear STEP; unfold stepInv;
 unfold S, stepInv, sameExec, sameEnv, exec, StateToMove, StateBeing, StateDone;
 intros [P [SD [TT TB]]].
(* Nop *)
simpl; auto.
(* Start *)
apply (noTmpLast_push m t1 t2); auto.
(* Push *)
apply (noTmpLast_push (d, r) t1 t2); auto.
(* Loop *)
generalize TB; rewrite <- app_cons; rewrite <- app_cons.
apply noTmpLast_tmpLast.
(* Pop *)
apply (noTmpLast_pop (sn, dn)); auto.
(* Last *)
auto.
Qed.
(* *)
 
Lemma step_inv: forall (r1 r2 : State), step r1 r2 -> stepInv r1 ->  stepInv r2.
Proof.
intros.
unfold stepInv.
split.
apply (step_inv_path r1 r2); auto.
split.
apply (step_inv_simpleDest r1 r2); auto.
split.
apply (step_inv_noTmp r1 r2); auto.
apply (step_inv_noTmpLast r1 r2); auto.
Qed.
(* *)
 
Definition step_NF (r : State) : Prop := ~ (exists s : State , step r s ).
(* * Transitive closure of step
   *)
 
Inductive stepp : State -> State ->  Prop :=
  stepp_refl: forall (r : State),  stepp r r
 | stepp_trans:
     forall (r1 r2 r3 : State), step r1 r2 -> stepp r2 r3 ->  stepp r1 r3 .
Hint Resolve stepp_refl stepp_trans .
 
Lemma stepp_transitive:
 forall (r1 r2 r3 : State), stepp r1 r2 -> stepp r2 r3 ->  stepp r1 r3.
Proof.
intros; induction H as [r|r1 r2 r0 H H1 HrecH]; eauto.
Qed.
 
Lemma step_stepp: forall (s1 s2 : State), step s1 s2 ->  stepp s1 s2.
Proof.
eauto.
Qed.
 
Lemma stepp_inv:
 forall (r1 r2 : State), stepp r1 r2 -> stepInv r1 ->  stepInv r2.
Proof.
intros; induction H as [r|r1 r2 r3 H H1 HrecH]; auto.
apply HrecH; auto.
apply (step_inv r1 r2); auto.
Qed.
 
Lemma stepp_sameExec:
 forall (r1 r2 : State), stepp r1 r2 -> stepInv r1 ->  sameExec r1 r2.
Proof.
intros; induction H as [r|r1 r2 r3 H H1 HrecH].
unfold sameExec, sameEnv; intros; auto.
lapply HrecH.
cut (sameExec r1 r2).
unfold sameExec, sameEnv.
intros.
Use (get (exec r1 e) r = get (exec r2 e) r); auto.
apply (step_sameExec r1); auto.
apply (step_inv r1); auto.
Qed.
(* * A step more deterministic
   *)
 
Inductive dstep : State -> State ->  Prop :=
  dstep_nop:
    forall (r : Reg) (t l : Moves),  dstep (S ((r, r) :: t) nil l) (S t nil l)
 | dstep_start:
     forall (t l : Moves) (s d : Reg),
     s <> d ->  dstep (S ((s, d) :: t) nil l) (S t ((s, d) :: nil) l)
 | dstep_push:
     forall (t1 t2 b l : Moves) (s d r : Reg),
     noRead t1 d ->
      dstep
       (S (t1 ++ ((d, r) :: t2)) ((s, d) :: b) l)
       (S (t1 ++ t2) ((d, r) :: ((s, d) :: b)) l)
 | dstep_pop_loop:
     forall (t b l : Moves) (s d r0 : Reg),
     noRead t r0 ->
      dstep
       (S t ((s, r0) :: (b ++ ((r0, d) :: nil))) l)
       (S t (b ++ ((tmp, d) :: nil)) ((s, r0) :: ((r0, tmp) :: l)))
 | dstep_pop:
     forall (t b l : Moves) (s0 d0 sn dn : Reg),
     noRead t dn ->
     dn <> s0 ->
      dstep
       (S t ((sn, dn) :: (b ++ ((s0, d0) :: nil))) l)
       (S t (b ++ ((s0, d0) :: nil)) ((sn, dn) :: l))
 | dstep_last:
     forall (t l : Moves) (s d : Reg),
     noRead t d ->  dstep (S t ((s, d) :: nil) l) (S t nil ((s, d) :: l)) .
Hint Resolve dstep_nop dstep_start dstep_push .
Hint Resolve dstep_pop_loop dstep_pop dstep_last .
 
Lemma dstep_step:
 forall (r1 r2 : State), dstep r1 r2 -> stepInv r1 ->  stepp r1 r2.
Proof.
intros r1 r2 DS; inversion_clear DS; intros SI.
(* nop *)
Use ((r, r) :: t = nil ++ ((r, r) :: t)).
eauto.
(* start *)
Use ((s, d) :: t = nil ++ ((s, d) :: t)).
eauto.
(* push *)
eauto.
(* pop_loop *)
apply
 (stepp_trans
   (S t ((s, r0) :: (b ++ ((r0, d) :: nil))) l)
   (S t ((s, r0) :: (b ++ ((tmp, d) :: nil))) ((r0, tmp) :: l))
   (S t (b ++ ((tmp, d) :: nil)) ((s, r0) :: ((r0, tmp) :: l)))).
auto.
apply step_stepp.
apply step_pop; auto.
unfold stepInv in SI |-; generalize SI; intros [X [Y [Z T]]].
generalize T; unfold S, StateBeing, noTmpLast.
case (b ++ ((r0, d) :: nil)); auto.
intros m l0 [R1 [OK PP]]; auto.
(* pop *)
eauto.
(* last *)
eauto.
Qed.
 
Lemma dstep_inv:
 forall (r1 r2 : State), dstep r1 r2 -> stepInv r1 ->  stepInv r2.
Proof.
intros.
apply (stepp_inv r1 r2); auto.
apply dstep_step; auto.
Qed.
 
Inductive dstepp : State -> State ->  Prop :=
  dstepp_refl: forall (r : State),  dstepp r r
 | dstepp_trans:
     forall (r1 r2 r3 : State), dstep r1 r2 -> dstepp r2 r3 ->  dstepp r1 r3 .
Hint Resolve dstepp_refl dstepp_trans .
 
Lemma dstepp_stepp:
 forall (s1 s2 : State), stepInv s1 -> dstepp s1 s2 ->  stepp s1 s2.
Proof.
intros; induction H0 as [r|r1 r2 r3 H0 H1 HrecH0]; auto.
apply (stepp_transitive r1 r2 r3); auto.
apply dstep_step; auto.
apply HrecH0; auto.
apply (dstep_inv r1 r2); auto.
Qed.
 
Lemma dstepp_sameExec:
 forall (r1 r2 : State), dstepp r1 r2 -> stepInv r1 ->  sameExec r1 r2.
Proof.
intros; apply stepp_sameExec; auto.
apply dstepp_stepp; auto.
Qed.
 
End pmov.
